@@ -7,6 +7,12 @@
 // files in the project carrying such notice may not be copied, modified,
 // or distributed except according to those terms.
 
+#![deny(warnings)]
+#![doc(test(attr(deny(warnings))))]
+#![doc(test(attr(allow(dead_code))))]
+#![doc(test(attr(allow(unused_variables))))]
+#![doc(test(attr(allow(unused_macros))))]
+
 //! This crate provides several macros for deriving some useful methods for unitary enums
 //! (*i.e.* enums where variants do not have payloads) and unary enums.
 //!
@@ -31,7 +37,6 @@
 //! ```
 //! can also be written as
 //! ```rust
-//! # use macro_attr_2018::macro_attr;
 //! # use enum_derive_2018::IterVariants;
 //! #[derive(Copy, Clone, Debug)]
 //! enum ItAintRight { BabeNo, NoNo, BoyBoy }
@@ -40,12 +45,12 @@
 //! # fn main() {}
 //! ```
 
-#![deny(warnings)]
-
-#![cfg_attr(not(feature = "std"), no_std)]
-#[cfg(feature = "std")]
+#![cfg_attr(not(feature="std"), no_std)]
+#[cfg(feature="std")]
 extern crate core;
 
+#[doc(hidden)]
+pub use core::compile_error as std_compile_error;
 #[doc(hidden)]
 pub use core::convert::From as std_convert_From;
 #[doc(hidden)]
@@ -75,7 +80,7 @@ macro_rules! enum_derive_util {
     (@first_expr $head:expr $(, $($($tail:expr),+ $(,)?)?)?) => {$head};
 
     (
-        @collect_unitary_variants ($callback:ident { $($args:tt)* }),
+        @collect_unitary_variants ($callback:path { $($args:tt)* }),
         ($(,)*) -> ($($var_names:ident,)*)
     ) => {
         $crate::enum_derive_util! {
@@ -108,11 +113,11 @@ macro_rules! enum_derive_util {
         @collect_unitary_variants ($name:ident),
         ($var:ident $_struct:tt, $($tail:tt)*) -> ($($var_names:tt)*)
     ) => {
-        const _error: () = "cannot parse unitary variants from enum with non-unitary variants";
+        $crate::std_compile_error!("cannot parse unitary variants from enum with non-unitary variants");
     };
 
     (
-        @collect_unary_variants ($callback:ident { $($args:tt)* }),
+        @collect_unary_variants ($callback:path { $($args:tt)* }),
         ($(,)*) -> ($($out:tt)*)
     ) => {
         $crate::enum_derive_util! {
@@ -145,11 +150,12 @@ macro_rules! enum_derive_util {
         @collect_unary_variants ($name:ident),
         ($var:ident $_struct:tt, $($tail:tt)*) -> ($($_out:tt)*)
     ) => {
-        const _error: () = "cannot parse unary variants from enum with non-unary tuple variants";
+        $crate::std_compile_error!("cannot parse unary variants from enum with non-unary tuple variants");
     };
 }
 
-/// Derives `iter_variants()` for an unitary enum, which returns an iterator over the variants of the enum in lexical order.
+/// Derives `iter_variants()` for an unitary enum,
+/// which returns an iterator over the variants of the enum in lexical order.
 ///  
 /// The argument is the name of the iterator type that will be generated:
 /// ```rust
@@ -169,11 +175,18 @@ macro_rules! IterVariants {
     (($itername:ident) $vis:vis enum $name:ident { $($body:tt)* }) => {
         $crate::enum_derive_util! {
             @collect_unitary_variants
-            (IterVariants { @expand ($vis) $itername, $name }),
+            ($crate::IterVariants_impl { @expand ($vis) $itername, $name }),
             ($($body)*,) -> ()
         }
     };
+    (() $vis:vis enum $name:ident { $($body:tt)* }) => {
+        $crate::std_compile_error!("IterVariants! requires iterator name parameter");
+    };
+}
 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! IterVariants_impl {
     (
         @expand ($vis:vis) $itername:ident, $name:ident ()
     ) => {
@@ -208,7 +221,7 @@ macro_rules! IterVariants {
     ) => {
         $crate::enum_derive_util! { @as_item $vis struct $itername($crate::std_option_Option<$name>); }
 
-        $crate::IterVariants! { @iter ($itername, $name), ($($var_names,)*) -> () () (0usize) }
+        $crate::IterVariants_impl! { @iter ($itername, $name), ($($var_names,)*) -> () () (0usize) }
 
         $crate::enum_derive_util! {
             @as_item
@@ -253,7 +266,7 @@ macro_rules! IterVariants {
     (
         @iter ($itername:ident, $name:ident), ($a:ident, $b:ident, $($rest:tt)*) -> ($($next_body:tt)*) ($($size_body:tt)*) ($($count:tt)*)
     ) => {
-        $crate::IterVariants! {
+        $crate::IterVariants_impl! {
             @iter ($itername, $name), ($b, $($rest)*)
             -> (
                 $($next_body)*
@@ -270,7 +283,7 @@ macro_rules! IterVariants {
     (
         @iter ($itername:ident, $name:ident), ($a:ident,) -> ($($next_body:tt)*) ($($size_body:tt)*) ($($count:tt)*)
     ) => {
-        $crate::IterVariants! {
+        $crate::IterVariants_impl! {
             @iter ($itername, $name), ()
             -> (
                 $($next_body)*
@@ -306,11 +319,18 @@ macro_rules! IterVariantNames {
     (($itername:ident) $vis:vis enum $name:ident { $($body:tt)* }) => {
         $crate::enum_derive_util! {
             @collect_unitary_variants
-            (IterVariantNames { @expand ($vis) $itername, $name }),
+            ($crate::IterVariantNames_impl { @expand ($vis) $itername, $name }),
             ($($body)*,) -> ()
         }
     };
+    (() $vis:vis enum $name:ident { $($body:tt)* }) => {
+        $crate::std_compile_error!("IterVariantNames! requires iterator name parameter");
+    };
+}
 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! IterVariantNames_impl {
     (
         @expand ($vis:vis) $itername:ident, $name:ident ()
     ) => {
@@ -345,7 +365,7 @@ macro_rules! IterVariantNames {
     ) => {
         $crate::enum_derive_util! { @as_item $vis struct $itername($crate::std_option_Option<$name>); }
 
-        $crate::IterVariantNames! { @iter ($itername, $name), ($($var_names,)*) -> () () (0usize) }
+        $crate::IterVariantNames_impl! { @iter ($itername, $name), ($($var_names,)*) -> () () (0usize) }
 
         $crate::enum_derive_util! {
             @as_item
@@ -391,7 +411,7 @@ macro_rules! IterVariantNames {
     (
         @iter ($itername:ident, $name:ident), ($a:ident, $b:ident, $($rest:tt)*) -> ($($next_body:tt)*) ($($size_body:tt)*) ($($count:tt)*)
     ) => {
-        $crate::IterVariantNames! {
+        $crate::IterVariantNames_impl! {
             @iter ($itername, $name), ($b, $($rest)*)
             -> (
                 $($next_body)*
@@ -409,7 +429,7 @@ macro_rules! IterVariantNames {
     (
         @iter ($itername:ident, $name:ident), ($a:ident,) -> ($($next_body:tt)*) ($($size_body:tt)*) ($($count:tt)*)
     ) => {
-        $crate::IterVariantNames! {
+        $crate::IterVariantNames_impl! {
             @iter ($itername, $name), ()
             -> (
                 $($next_body)*
@@ -425,17 +445,22 @@ macro_rules! IterVariantNames {
     };
 }
 
-/// Derives `next_variant(&self)` for an unitary enum, which returns the next variant, or `None` when called for the last.
+/// Derives `next_variant(&self)` for an unitary enum,
+/// which returns the next variant, or `None` when called for the last.
 #[macro_export]
 macro_rules! NextVariant {
     (() $vis:vis enum $name:ident { $($body:tt)* }) => {
         $crate::enum_derive_util! {
             @collect_unitary_variants
-            (NextVariant { @expand ($vis) $name }),
+            ($crate::NextVariant_impl { @expand ($vis) $name }),
             ($($body)*,) -> ()
         }
     };
+}
 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! NextVariant_impl {
     (
         @expand ($vis:vis) $name:ident ()
     ) => {
@@ -458,7 +483,7 @@ macro_rules! NextVariant {
             impl $name {
                 #[allow(dead_code)]
                 $vis fn next_variant(&self) -> $crate::std_option_Option<$name> {
-                    $crate::NextVariant!(@arms ($name, self), ($($var_names)*) -> ())
+                    $crate::NextVariant_impl!(@arms ($name, self), ($($var_names)*) -> ())
                 }
             }
         }
@@ -479,7 +504,7 @@ macro_rules! NextVariant {
     (
         @arms ($name:ident, $self_:expr), ($a:ident $b:ident $($rest:tt)*) -> ($($body:tt)*)
     ) => {
-        $crate::NextVariant! {
+        $crate::NextVariant_impl! {
             @arms ($name, $self_), ($b $($rest)*)
             -> (
                 $($body)*
@@ -495,11 +520,15 @@ macro_rules! PrevVariant {
     (() $vis:vis enum $name:ident { $($body:tt)* }) => {
         $crate::enum_derive_util! {
             @collect_unitary_variants
-            (PrevVariant { @expand ($vis) $name }),
+            ($crate::PrevVariant_impl { @expand ($vis) $name }),
             ($($body)*,) -> ()
         }
     };
+}
 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! PrevVariant_impl {
     (
         @expand ($vis:vis) $name:ident ()
     ) => {
@@ -522,7 +551,7 @@ macro_rules! PrevVariant {
             impl $name {
                 #[allow(dead_code)]
                 $vis fn prev_variant(&self) -> $crate::std_option_Option<$name> {
-                    $crate::PrevVariant!(@arms ($name, self), ($crate::std_option_Option::None, $($var_names)*) -> ())
+                    $crate::PrevVariant_impl!(@arms ($name, self), ($crate::std_option_Option::None, $($var_names)*) -> ())
                 }
             }
         }
@@ -543,7 +572,7 @@ macro_rules! PrevVariant {
     (
         @arms ($name:ident, $self_:expr), ($prev:expr, $a:ident $($rest:tt)*) -> ($($body:tt)*)
     ) => {
-        $crate::PrevVariant! {
+        $crate::PrevVariant_impl! {
             @arms ($name, $self_), ($crate::std_option_Option::Some($name::$a), $($rest)*)
             -> (
                 $($body)*
@@ -561,11 +590,15 @@ macro_rules! EnumDisplay {
     (() $vis:vis enum $name:ident { $($body:tt)* }) => {
         $crate::enum_derive_util! {
             @collect_unitary_variants
-            (EnumDisplay { @expand $name }),
+            ($crate::EnumDisplay_impl { @expand $name }),
             ($($body)*,) -> ()
         }
     };
+}
 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! EnumDisplay_impl {
     (
         @expand $name:ident ()
     ) => {
@@ -586,7 +619,7 @@ macro_rules! EnumDisplay {
             @as_item
             impl $crate::std_fmt_Display for $name {
                 fn fmt(&self, f: &mut $crate::std_fmt_Formatter) -> $crate::std_fmt_Result {
-                    $crate::EnumDisplay!(@arms ($name, self, f), ($($var_names)*) -> ())
+                    $crate::EnumDisplay_impl!(@arms ($name, self, f), ($($var_names)*) -> ())
                 }
             }
         }
@@ -607,7 +640,7 @@ macro_rules! EnumDisplay {
     (
         @arms ($name:ident, $self_:expr, $f:ident), ($a:ident $b:ident $($rest:tt)*) -> ($($body:tt)*)
     ) => {
-        $crate::EnumDisplay! {
+        $crate::EnumDisplay_impl! {
             @arms ($name, $self_, $f), ($b $($rest)*)
             -> (
                 $($body)*
@@ -623,11 +656,15 @@ macro_rules! EnumFromStr {
     (() $vis:vis enum $name:ident { $($body:tt)* }) => {
         $crate::enum_derive_util! {
             @collect_unitary_variants
-            (EnumFromStr { @expand ($vis) $name }),
+            ($crate::EnumFromStr_impl { @expand ($vis) $name }),
             ($($body)*,) -> ()
         }
     };
+}
 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! EnumFromStr_impl {
     (
         @expand ($vis:vis) $name:ident ()
     ) => {
@@ -652,7 +689,7 @@ macro_rules! EnumFromStr {
                 type Err = $crate::ParseEnumError;
 
                 fn from_str(s: &str) -> Result<Self, Self::Err> {
-                    $crate::EnumFromStr!(@arms ($name, s), ($($var_names)*) -> ())
+                    $crate::EnumFromStr_impl!(@arms ($name, s), ($($var_names)*) -> ())
                 }
             }
         }
@@ -674,7 +711,7 @@ macro_rules! EnumFromStr {
     (
         @arms ($name:ident, $s:ident), ($a:ident $b:ident $($rest:tt)*) -> ($($body:tt)*)
     ) => {
-        $crate::EnumFromStr! {
+        $crate::EnumFromStr_impl! {
             @arms ($name, $s), ($b $($rest)*)
             -> (
                 $($body)*
@@ -695,23 +732,24 @@ impl core::fmt::Display for ParseEnumError {
 }
 
 #[cfg(feature="std")]
-impl std::error::Error for ParseEnumError {
-    fn description(&self) -> &str {
-        "provided string did not match any enum variant"
-    }
-}
+impl std::error::Error for ParseEnumError { }
 
-/// Derives [`From<T>`](core::convert::From) for each variant's payload, assuming all variants are unary.
+/// Derives [`From<T>`](core::convert::From) for each variant's payload,
+/// assuming all variants are unary.
 #[macro_export]
 macro_rules! EnumFromInner {
     (() $vis:vis enum $name:ident { $($body:tt)* }) => {
         $crate::enum_derive_util! {
             @collect_unary_variants
-            (EnumFromInner { @expand $name }),
+            ($crate::EnumFromInner_impl { @expand $name }),
             ($($body)*,) -> ()
         }
     };
+}
 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! EnumFromInner_impl {
     (
         @expand $name:ident ($($var_names:ident($var_tys:ty),)*)
     ) => {
@@ -751,21 +789,33 @@ macro_rules! EnumInnerAsTrait {
     ($arg:tt $vis:vis enum $name:ident { $($body:tt)* }) => {
         $crate::enum_derive_util! {
             @collect_unary_variants
-            (EnumInnerAsTrait { @expand $arg, $name, }),
+            ($crate::EnumInnerAsTrait_impl { @expand $arg, $name, }),
             ($($body)*,) -> ()
         }
     };
+}
 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! EnumInnerAsTrait_impl {
     (
         @expand ($vis:vis $fn_name:ident -> &mut $tr:ty), $($tail:tt)*
     ) => {
-        $crate::EnumInnerAsTrait! { @expand_inner ($vis), $fn_name, (mut), $tr, $($tail)* }
+        $crate::EnumInnerAsTrait_impl! { @expand_inner ($vis), $fn_name, (mut), $tr, $($tail)* }
     };
 
     (
         @expand ($vis:vis $fn_name:ident -> &$tr:ty), $($tail:tt)*
     ) => {
-        $crate::EnumInnerAsTrait! { @expand_inner ($vis), $fn_name, (), $tr, $($tail)* }
+        $crate::EnumInnerAsTrait_impl! { @expand_inner ($vis), $fn_name, (), $tr, $($tail)* }
+    };
+
+    (
+        @expand ($($t:tt)*), $($tail:tt)*
+    ) => {
+        $crate::std_compile_error!(
+            "invalid EnumInnerAsTrait! parameter, the allowed form is '$vis:vis $fn_name:ident -> & $(mut)? $tr:ty'"
+        );
     };
 
     (
